LOGS_DATA = "logs-data"

ICON_DARK_MODE = "dark-mode-icon"
BTN_DARK_MODE = "dark-mode-toggle"

LOADING_UPLOAD = "loading-upload"
LOADING_DEPTHMAP = "loading-depthmap"
LOADING_GENERATE_SLICE = "loading-generate-slice"
LOADING_GENERATE_INPAINTING = "generate-inpainting"
LOADING_GLTF = "gltf-loading"
LOADING_ANIMATION = "animation-loading"

CTR_DEPTH_MAP = "depth-map-container"
BTN_GENERATE_DEPTHMAP = "generate-depthmap-button"
DEPTHMAP_OUTPUT = "gen-depthmap-output"
ANIMATION_OUTPUT = "gen-animation-output"

CANVAS_DATA = "canvas-data"
CANVAS_MASK_DATA = "canvas-mask-data"
CANVAS_PAINT = "canvas-paint"

STORE_SELECTED_SLICE = "selected-slice"
STORE_IGNORE = "store-ignore"
STORE_GENERATE_SLICE = "generate-slice-request"
STORE_UPDATE_SLICE = "update-slice-request"
STORE_INPAINTING = "inpainting-request"
STORE_APPSTATE_FILENAME = "application-state-filename"
STORE_RESTORE_STATE = "restore-state"
STORE_RECT_DATA = "rect-data"
STORE_BOUNDING_BOX = "bounding-box"
STORE_TRIGGER_GEN_DEPTHMAP = "trigger-generate-depthmap"
STORE_TRIGGER_UPDATE_DEPTHMAP = "trigger-update-depthmap"
STORE_UPDATE_THRESHOLD_CONTAINER = "update-thresholds-container"
STORE_CLICKED_POINT = "clicked-point"
STORE_CLEAR_PREVIEW = "clear-preview"

TEXT_POSITIVE_PROMPT = "positive-prompt"
TEXT_NEGATIVE_PROMPT = "negative-prompt"

CTR_INPUT_IMAGE = "input-image-container"
IMAGE = "image"
CANVAS = "canvas"
PREVIEW_CANVAS = "preview-canvas"
UPLOAD_IMAGE = "upload-image"
UPLOAD_STATE = "upload-state"
UPLOAD_SLICE = "slice-upload"
BTN_SAVE_STATE = "save-state"

SLIDER_INPAINT_STRENGTH = "inpaint-strength"
SLIDER_INPAINT_GUIDANCE = "inpaint-guidance"
SLIDER_NUM_SLICES = "num-slices-slider"
SLIDER_MASK_PADDING = "mask-padding-slider"
SLIDER_MASK_BLUR = "mask-blur-slider"

SLIDER_CAMERA_DISTANCE = "camera-distance-slider"
SLIDER_MAX_DISTANCE = "max-distance-slider"
SLIDER_FOCAL_LENGTH = "focal-length-slider"
SLIDER_DISPLACEMENT = "displacement-slider"
SLIDER_NUM_FRAMES = "number-of-frames-slider"

DROPDOWN_DEPTH_MODEL = "depth-model-dropdown"
DROPDOWN_INPAINT_MODEL = "inpainting-model-dropdown"
DROPDOWN_MODE_SELECTOR = "mode-selector"

INPUT_EXTERNAL_SERVER = "external-server-address"
BTN_EXTERNAL_TEST_CONNECTION = "external-test-connection-button"

UPLOAD_COMFYUI_WORKFLOW = "comfyui-workflow-upload"
CTR_COMFYUI_WORKFLOW = "comfyui-workflow-container"
CTR_AUTOMATIC_CONFIG = "automatic-config-container"

INPUT_API_KEY = "api-key"
BTN_VALIDATE_API_KEY = "validate-api-key"
CTR_API_KEY = "api-key-container"

DOWNLOAD_IMAGE = "download-image"
DOWNLOAD_GLTF = "download-gltf"
DOWNLOAD_ANIMATION = "download-animation"

BTN_EXPORT_ANIMATION = "animation-export"

CTR_THRESHOLDS = "thresholds-container"
CTR_SLICE_IMAGES = "slice-img-container"

BTN_GENERATE_INPAINTING = "generate-inpainting-button"
BTN_FILL_INPAINTING = "fill-inpainting-button"
BTN_ENHANCE = "enhance-button"
BTN_ERASE_INPAINTING = "erase-inpainting-button"
BTN_APPLY_INPAINTING = "apply-inpainting-button"
CTR_INPAINTING_OUTPUT = "inpainting-output"
CTR_INPAINTING_IMAGES = "inpainting-img-container"
CTR_INPAINTING_DISPLAY = "inpainting-image-display"
CTR_INPAINTING_COLUNM = "inpainting-column"

ID_INPAINTING_IMAGE = "inpainting-image"

ID_SLICE_DEPTH_DISPLAY = "depth-display"
ID_SLICE_OVERLAY = "slicer-overlay"
INPUT_SLICE_DEPTH = "depth-input"

BTN_GENERATE_SLICE = "generate-slice-button"
BTN_BALANCE_SLICE = "balance-slice-button"
BTN_CREATE_SLICE = "create-slice-button"
BTN_DELETE_SLICE = "delete-slice-button"
BTN_ADD_SLICE = "add-to-slice-button"
BTN_REMOVE_SLICE = "remove-from-slice-button"
BTN_COPY_SLICE = "copy-button"
BTN_PASTE_SLICE = "paste-button"

BTN_CLEAR_CANVAS = "clear-canvas"
BTN_ERASE_MODE = "erase-mode-canvas"
BTN_LOAD_CANVAS = "load-canvas"

CTR_CANVAS_BUTTONS = "canvas-buttons"
NAV_ZOOM_OUT = "nav-zoom-out"
NAV_UP = "nav-up"
NAV_DOWN = "nav-down"
NAV_LEFT = "nav-left"
NAV_RIGHT = "nav-right"
NAV_ZOOM_IN = "nav-zoom-in"
NAV_RESET = "nav-reset"

CTR_SEG_BUTTONS = "segmentation-buttons"
SEG_TOGGLE_CHECKERBOARD = "toggle-checkerboard"
SEG_INVERT_MASK = "invert-mask"
SEG_FEATHER_MASK = "feather-mask"
SEG_MULTI_POINT = "multi-point"
SEG_MULTI_COMMIT = "multi-commit"

CTR_PROGRESS_BAR = "progress-bar-container"
PROGRESS_INTERVAL = "progress-interval"

CTR_GLTF_OUTPUT = "gen-gltf-output"
BTN_GLTF_CREATE = "gltf-create"
BTN_GLTF_EXPORT = "gltf-export"
BTN_UPSCALE_TEXTURES = "upscale-textures"

CTR_MODEL_VIEWER = "model-viewer-container"
IFRAME_MODEL_VIEWER = "model-viewer"

CTR_HELP_WINDOW = "help-window"
STORE_CURRENT_TAB = "current-tab"

CHECKLIST_DOF = "toggle-dof-support"
CHECKLIST_REGION_OF_INTEREST = "toggle-region-of-interest"
