module.exports = {
  content: [
    './parallax_maker/assets/**/*.js', './parallax_maker/*.py',
    '!./parallax_maker/test_*.py', '!./parallax_maker/stabilityai.py', '!./parallax_maker/htmlcov/**',
  ],
  darkMode: 'selector',
  theme: {
    extend: {},
  },
  variants: {
    extend: {},
  },
  plugins: [],
}
